LOCALS @@
TITLE   Turbo Pop Version 2.0a
SUBTTL  Copyright (c)1988 Ross Neilson Wentworth, All Rights Reserved

; Requires TASM 1.01 to assembler properly.  Minor modifications
; required to assembler with MASM 5.1.

; This code is Copyright (c)1988,89 by Ross Neilson Wentworth
; All Rights Reserved
;
; No part of this code may be reproduced, modified, sold, or used
; as an incentive to buy without the expressed written consent of
; the author.


; Routine Quick Reference
;
;   EnterPopUp          Saves and sets the stack then calls a high-
;                       level routine.
;   InitializePopUp     Sets the low-level interrupt vectors.
;   Installed           Sees if the program is already loaded.
;   new05Int            Optional print-screen trap.
;   new08int            Called at each timer tick.  Checks to see if we
;                       want to pop up and whether it is safe to do so.
;   new09int            Checks for our hotkey combo at each keypress.
;   new10Int            Optional video BIOS trap.
;   new13int            Keeps track of when the disk is busy.
;   new16int            Prevents popups from being locked out while ours
;                       is active.
;   new28int            Interrupt routine that is called whenever it is safe
;                       to interrupt DOS.
;   new2Fint            Intercepts the Multiplex interupt.

TRUE     equ    1
FALSE    equ    0

tries_allowed  EQU  18d

TRAPINT05  EQU  FALSE    ; change to TRUE to trap print-screens
TRAPINT10  EQU  FALSE    ; change to TRUE to trap video BIOS
TRAPINT16  EQU  TRUE     ; prevent others from being locked out

.MODEL TPASCAL

.DATA

EXTRN prefixseg:WORD

.CODE

       ASSUME CS:@CODE

PUBLIC InitializePopUp,Installed

EXTRN callpopup:FAR,ReleaseProgram:FAR

;----------------------------------------------------------------------------
; Miscellaneous data
;----------------------------------------------------------------------------
hotkeyone  LABEL  WORD
scancode1     DB  0    ; scan code of our hotkey
shiftstatus1  DB  0    ; shift key status (alt,ctrl, etc. )

hotkeytwo  LABEL WORD  ; alternate hotkey
scancode2     DB  0
shiftstatus2  DB  0

program_status DB  0    ; true if the program is active
request_flag   DB  0    ; true if we want to pop-up
notsafe        DB  0    ; non-zero if unsafe to pop up
multiplexID    DB  0    ; unique program ID code
tries          DB  0    ; number of timer ticks we've tried to pop up

stack_pointer  DW  0    ; holds our program's stack pointer
stack_segment  DW  0    ; holds our program's stack segment
old_pointer    DW  0    ; holds the interrupted stack pointer
old_stack      DW  0    ; holds the interrupted stack segment

dos_busy_flag  DD  0    ; points to DOS's busy flag
SigAddr        DD  0    ; signature string address

; Follows is the interrupt service routine (ISR) installation table.
; Each interrupt will have three entries, the interrupt number, the
; original interrupt vector, and the offset of the ISR.  The entire
; table is terminated with a zero byte (this disallows the automatic
; installation of interrupt 00h).  There are two ISR's that are optional.
; One traps the print-screen interrupt (05h) and the other traps the video
; BIOS interrupt (10h).  By default, these traps are not used.  If you wish
; to include them in the interrupt trapping list you must change the EQU's
; above to TRUE.  TRAPINT05 is the EQU for trapping the print-screen
; interrupt and TRAPINT10 is the EQU for trapping the video BIOS.

;----------------------------------------------------------------------------
; Interrupt Vector Table
;----------------------------------------------------------------------------

Vectors  LABEL   BYTE

IF TRAPINT05               ; optional interrupt trap

               DB  05h     ; print screen vector number
old05int       DD  ?       ; the original print screen interrupt vector
               DW  offset new05int
ENDIF

               DB  08h     ; timer tick
old08int       DD  ?       ; the original timer tick interrupt vector
               DW  offset new08int

               DB  09h     ; keyboard vector number
old09int       DD  ?       ; the original keyboard interrupt vector
               DW  offset new09int

IF TRAPINT10               ; optional interrupt trap

               DB  10h     ; video BIOS vector number
old10int       DD  ?       ; the original video BIOS interrupt vector
               DW  offset new10int
ENDIF

               DB  13h     ; disk i/o vector number
old13int       DD  ?       ; the original disk interrupt vector
               DW  offset new13int

IF TRAPINT16   ; this trap should not be used if you are using TPCRT
               ; from the Turbo Professional package

               DB  16h     ; keyboard vector number
old16int       DD  ?       ; the original keyboard interrupt vector
               DW  offset new16int
ENDIF

               DB  28h     ; backprocess vector number
old28int       DD  ?       ; the original backprocess interrupt vector
               DW  offset new28int

               DB  2Fh     ; multiplex interrupt number
old2Fint       DD  ?       ; the original multiplex interrupt vector
               DW  offset new2Fint

               DB  0       ; end of table

;----------------------------------------------------------------------------
; Print-Screen Interrupt Service Routine - 05h
;----------------------------------------------------------------------------
; This is an optional interrupt trap.  Change the value of TRAPINT05 to
; TRUE if you want it compiled.  This one traps the PRINT SCREEN interrupt
; and prevents popping up in the middle of a screen print.

IF  trapInt05

new05int PROC  FAR

        inc     cs:[notsafe]    ; increment status flag

        pushf
        cli
        call    cs:[old05int]   ; chain to original INT 05h handler

        pushf                   ; preserve flags
        dec     cs:[notsafe]    ; decrement status flag
        popf                    ; restore flags

        sti                     ; enable interrupts
        ret     2               ; return with flags

new05int ENDP

ENDIF

;----------------------------------------------------------------------------
; Keyboard Interrupt Service Routine - 09h
;----------------------------------------------------------------------------
; Every keystroke is checked through here for our hot_key combination.
; If our's is pressed, a service request flag is set to TRUE.  If the
; program is already active we let the keystroke by so that conflicting
; hot-keys will work.

new09int PROC  FAR

        sti
        pushf                             ; save the flags
        push    ax
        cmp     cs:[program_status],FALSE ; are we already active?
        jne     L200                      ; yes, so chain to original vector
        cmp     cs:[request_flag],FALSE   ; do we already want serveice?
        je      L201                      ; no, so check for hot-key
L200:
        pop     ax
        popf
        jmp     cs:[old09int]          ; chain to original vector

L201:
        push   es
        xor    ax,ax
        mov    es,ax                   ; point ES to low memory
        mov    ah,es:[0417h]           ; get the keyboard flags
        pop    es

        and     ah,0Fh                 ; mask off unused keyflag bits
        in      al,60h                 ; get the keystroke
        cmp     ax,cs:[HotKeyOne]      ; is it hot key 1 ?
        je      L202                   ; yes!  so jump
        cmp     ax,cs:[HotKeyTwo]      ; is it hot key 2 ?
        jne     L200                   ; no, jump back and exit
L202:
        cli                            ; turn off interrupts
        mov     al,20h                 ; say interrupt is finished
        out     20h,al

        in      al,61h                 ; trash the keystroke
        mov     ah,al
        or      al,80h
        out     61h,al
        jmp     @@1                    ; jump to nowhere for delay
@@1:
        mov     al,ah
        out     61h,al
        mov     cs:[tries],tries_allowed  ; set number of tries to make
        mov     cs:[request_flag],TRUE ; say service is desired
        pop     ax                     ; restore used register
        popf                           ; restore flags
        iret                           ; all done

new09int ENDP

;----------------------------------------------------------------------------
; Video BIOS Interrupt Service Routine - 10h
;----------------------------------------------------------------------------
; This is an optional interrupt trap.  Change the value of TRAPINT10 to
; TRUE if you want it compiled.  This one traps the video BIOS interrupt
; and prevents popping up in the middle of any screen activity.

IF  trapInt10

new10int PROC  FAR

        inc     cs:[notsafe]    ; increment status flag

        pushf
        cli
        call    cs:[old10int]   ; chain to original INT 10h handler

        pushf                   ; preserve flags
        dec     cs:[notsafe]    ; decrement status flag
        popf                    ; restore flags

        sti                     ; enable interrupts
        ret     2               ; return with flags

new10int ENDP

ENDIF

;----------------------------------------------------------------------------
; Disk I/O Interrupt Service Routine - 13h
;----------------------------------------------------------------------------
; At each disk call, increments an unsafe flag, finishes
; the disk call, then decrements the unsafe flag.  This
; prevents an unsafe "popup" during disk i/o.

new13int PROC FAR

        inc     cs:[notsafe]    ; increment status flag

        pushf                   ; chain to original INT 13h handler
        cli
        call    cs:[old13int]

        pushf                   ; preserve flags
        dec     cs:[notsafe]    ; decrement status flag
        popf                    ; restore flags

        sti                     ; enable interrupts
        ret     2               ; return with flags, this is to retain
                                ; any carry flag (error) that may
                                ; have occurred.

new13int ENDP

;----------------------------------------------------------------------------
; Keyboard interrupt - 16h
;----------------------------------------------------------------------------
; Traps the keyboard function to get a key, this is to prevent other resident
; programs from being locked out while ours is active.

IF TrapInt16

new16int PROC   FAR

        sti
        pushf
        push    ax
        cmp     ah,10h          ; enhanced keyboard get key?
        je      @@3             ; yes, so jump
        or      ah,ah           ; standard keyboard get key?
        jnz     skip            ; no, so jump to exit
@@3:
        int     28h             ; backprocess interrupt
        pop     ax              ; get back function
        push    ax              ; and save it again
        inc     ah              ; change to query for key (test key)
        pushf                   ; simulate INT 16h
        cli
        call    cs:[old16int]   ; check for a keystroke
        jz      @@3             ; repeat until key is pressed

skip:
        pop     ax
        popf
        cli
        jmp     cs:[old16int]

new16int ENDP

ENDIF

;----------------------------------------------------------------------------
; Multiplex Interrupt Service Routine - 2Fh
;----------------------------------------------------------------------------
; We use this ISR to check if a program is already loaded as well as to
; provide special routines needed to unload the program safely.
;
; function 0, returns 0FFh in AL to indicate the program is already loaded
; function 1, returns the vector table in DX:AX
; function 2, returns the module's code segment (CS) in AX
; function 3, releases the memory for this program
; function 4, return the segment:offset of the signature string in DX:AX
;
; Examine the UNHOOK unit to see how these are used

new2Fint PROC   FAR

        cmp     ah,cs:[MultiplexID]   ; is it our id number?
        je      @@4                   ; yes, go check function number
        jmp     cs:[old2Fint]         ; chain to original 2Fh interrupt vector
@@4:
        test    al,al                 ; function 0, loaded flag
        jnz     @@5
        mov     al,0FFh
        les     di,cs:[SigAddr]
        jmp     short L300
@@5:
        cmp     al,1                  ; function 1, return vector table pointer
        jne     @@6
        mov     dx,cs
        lea     ax,cs:[vectors]
        jmp     short L300
@@6:
        cmp     al,2                  ; function 2, return code segment
        jne     @@7
        mov     ax,cs
        jmp     short L300
@@7:
        cmp     al,3                  ; function 3, release program memory
        jne     @@08
        pushf
        call    ReleaseProgram
        jmp     short L300
@@08:
        cmp     al,4                  ; function 4, return signature address
        jne     L300
        mov     ax,word ptr cs:[SigAddr]
        mov     dx,word ptr cs:[SigAddr+2]
L300:
        iret

new2Fint ENDP

;----------------------------------------------------------------------------
; Timer Tick Interrupt Service Routine - 08h
;----------------------------------------------------------------------------
; At each timer tick, this routine sees if our routine wants service.
; If it does, and it's safe to run, it saves junk and calls our routine.

new08int PROC  FAR

        pushf
        cli
        call    cs:[old08int]
        cmp     cs:[request_flag],FALSE   ; request flag set?
        je      L404                      ; no, then exit
        push    di
        push    es                        ; save es and di
        les     di,cs:[dos_busy_flag]     ; address of DOS BUSY_FLAG in DI
        cmp     byte ptr es:[di],FALSE    ; DOS service currently active?
        pop     es
        pop     di                        ; clean up the stack
        jne     L400                      ; dos active, so wait
        cmp     cs:[notsafe],FALSE        ; is it safe to pop-up?
        je      EnterPopUp                ; yes, go for it
L400:
        dec     cs:[tries]
        jnz     L404                      ; jump if no timeout
        mov     cs:[request_flag],FALSE   ; give up

; buzz to indicate a timeout

        push    ax
        push    cx
        push    dx
        mov     dl,0Ah
        in      al,61h
        and     al,0FEh
L401:
        or      al,2
        out     61h,al
        mov     cx,1000h
L402:   loop    L402
        and     al,0FDh
        out     61h,al
        mov     cx,500h
@L403:  loop    @L403
        dec     dl
        jnz     L401
        pop     dx
        pop     cx
        pop     ax
L404:
        iret

new08int ENDP

;----------------------------------------------------------------------------
; Backprocess Interrupt Service Routine - 28h
;----------------------------------------------------------------------------
; Whenever it is safe for a program to interrupt, DOS issues an
; INT 28h.  This normally occurs while sitting at the command line.
; This routine checks to see if our program wants service and pops
; it up if it does.

new28int PROC FAR

        pushf                           ; call the original interrupt routine
        cli
        call    cs:[old28int]

        cmp     cs:[request_flag],FALSE ; do we want to pop up?
        jne     EnterPopUp              ; yes, go pop up
        iret                            ; no. exit

new28int ENDP

;----------------------------------------------------------------------------
; Setup For Entering Application
;----------------------------------------------------------------------------
; Sets SS, and SP, then calls the popup program.
; After returning, it restores them.

EnterPopUp  PROC   NEAR

        mov     cs:[request_flag],FALSE  ; clear the pop up request flag
        mov     cs:[program_status],TRUE ; say program is active
        mov     cs:[old_stack],ss        ; save the stack segment
        mov     cs:[old_pointer],sp      ; save the stack pointer
        cli                              ; turn off interrupts
        mov     ss,cs:[stack_segment]    ; get our program's stack segment
        mov     sp,cs:[stack_pointer]    ; get our program's stack pointer
        sti                              ; turn interrupts back on

; Now we enter the high-level language portion of the program.  We've
; saved all the low-level information, such as registers, stack, etc.
; it is up to the high-level portion to do save and set the DTA, as
; well as the Critical Error interrupt vector and the Control-Break
; interrupt vector.  See the unit source POPUP.PAS for the exact
; details.

        pushf                            ; simulate an interrupt
        call    CallPopUp

; We're back from the popup application, now
; to return to the interrupted program.

; restore the stack segment and pointer

        cli                               ; old cpu bug require ints be off
        mov     ss,cs:[old_stack]
        mov     sp,cs:[old_pointer]
        sti

        mov     cs:[program_status],FALSE ; say program is not active
        iret                              ; all done

EnterPopUp  ENDP

;----------------------------------------------------------------------------
; Initialize memory and the vectors
;----------------------------------------------------------------------------
; We save some interrupt vectors and flags in the code segment for
; efficiency.  We don't want to be pushing and popping all of the
; registers at every single timer tick and key press.
;
; It's too bad there isn't any way of releasing the space used by the
; initialization code.  This is the reason the most efficient resident
; programs are written in entirely assembler, complete control of segment
; ordering.

InitializePopUp PROC NEAR USES DS, MultId:BYTE:2,KeyOne:WORD,KeyTwo:WORD

; save our hot key combo

        mov     ax,KeyOne              ; save hotkey one
        mov     cs:[HotKeyOne],ax
        mov     ax,KeyTwo              ; save hotkey two
        mov     cs:[HotKeyTwo],ax

; save the TSR's unique ID

        mov     al,multID              ; save the I.D. number
        mov     cs:[multiplexID],al

; save the stack segment and pointer

        mov     cs:[stack_segment],ss
        mov     cs:[stack_pointer],sp

; get the DOS busy flag address

        mov     ah,34h
        int     21h
        mov     cs:word ptr [dos_busy_flag+2],es
        mov     cs:word ptr [dos_busy_flag],bx

        push    cs
        pop     ds              ; point to the code segment

; set all of the interrupt vectors

        lea     si,vectors
        cld
@@9:
        lodsb                   ; get a vector number
        or      al,al           ; are we done
        jz      @@10            ; yes
        push    ax              ; save the interrupt number
        mov     ah,35h
        int     21h             ; get the old vector
        mov     [si],bx         ; save the offset
        mov     [si+2],es       ; save the segment
        pop     ax              ; get back interrupt number
        mov     ah,25h          ; set the interrupt vector
        mov     dx,[si+4]       ; get the routine offset assume this code segment
        int     21h
        add     si,6            ; point to next item in table
        jmp     short @@9       ; do it again
@@10:
        ret

InitializePopUp  ENDP

;----------------------------------------------------------------------------
; Checks for a previously installed program
;----------------------------------------------------------------------------
; Checks to see if the program is already loaded by using the Multiplex
; interrupt (2Fh).  It must do some convaluted crap to avoid problems
; under DOS 2.x.

Installed PROC  FAR USES DS,Signature:FAR PTR,MultId:BYTE:2

        lds     si,Signature                 ; save the signature string now
        mov     word ptr cs:[SigAddr],si
        mov     word ptr cs:[SigAddr+2],ds
        mov     ah,30h
        int     21h             ; get the DOS version
        cmp     al,2
        ja      L101            ; jump if version 3.0 or later
        mov     ax,352Fh        ; get vector for 2Fh
        int     21h
        mov     ax,es
        or      ax,bx           ; jump if current INT 2Fh vector ..
        jnz     L100            ; .. is nonzero

        mov     ax,0F000h       ; if INT 2Fh is NULL then we will ..
        mov     es,ax           ; .. find an IRET in the bios and ..
        mov     di,-1           ; .. point it to it.
        mov     cx,-1
        std
        mov     al,0CFh
        repne   scasb           ; find an IRET in the BIOS
        inc     di

        mov     dx,di
        push    ds
        push    es
        pop     ds
        mov     ax,252Fh
        int     21h             ; point 2Fh to an IRET in the BIOS
        pop     ds
        jmp     short L103      ; now go and install it
L100:
        mov     ax,0FF00h       ; look for PRINT.COM
        int     2Fh             ; if resident AH = print queue length
        cmp     ah,0FFh         ; otherwise AH is unchanged
        je      L101            ; use multiplex interrupt
        mov     al,1
        jmp     short L104      ; abort if PRINT.COM already installed
L101:
        mov     ah,MultId       ; get interrupt ID number
        xor     al,al
        int     2Fh             ; poll for the ID number
        test    al,al           ; installed already?
        jz      L103            ; jump if ok to install
        cmp     al,0FFh
        jne     L102            ; jump if not already installed
        lds     si,cs:[SigAddr]
        lodsb
        cmp     al,es:[di]
        jne     L105
        mov     cl,al
        xor     ch,ch
        inc     di
        repe    cmpsb
        je      L106
L105:
        mov     al,2
        jmp     short L104      ; error, number already used
L106:
        mov     al,3
        jmp     short L104      ; error already installed
L102:
        mov     al,4
        jmp     short L104      ; error, can't install
L103:
        xor     al,al
L104:
        xor     ah,ah
        ret

Installed ENDP

CODE ENDS

END
