TITLE  String Handling Routines
SUBTTL Copyright (c)1989 Ross Neilson Wentworth, All Rights Reserved

.MODEL TPASCAL
.CODE
LOCALS @@

PUBLIC uppercase,lowercase,locase,left,Right,NoLeadSpace,NoTrailSpace
PUBLIC Reverse,LastPos,PadLeft,PadRight,DelChar,ChangeChar
PUBLIC Chars,LongMonth,LongDay,WordCount,PosWord,NextPos,PrevPos
PUBLIC ShortMonth,ShortDay

; UpperCase    -  converts a string to upper case
; LowerCase    -  converts a string to lower case
; LoCase       -  converts a single character to lower case
; Left         -  returns the left n characters of a string
; Right        -  returns the right n characters of a string
; NoLeadSpace  -  Removes leading whitespace from a string
; noTrailSpace -  Removes trailing whitespace from a string
; Reverse      -  Returns a mirror image of the string
; LastPos      -  Returns the last occurrance of a character in a string
; NextPos      -  Returns the next occurance of a character in a string
; PrevPos      -  Returns the previous occurrance of a character in a string
; PadLeft      -  Inserts enough characters at the start of a string to
;                 make it a specified length
; PadRight     -  Adds enough characters to the end of the string to make
;                 it a specified length
; DelChar      -  Deletes all occurances of a specified character in the string
; ChangeChar   -  Replaces all occurances of a character with another
; Chars        -  Returns a string of n characters
; LongMonth    -  Returns the full name of a month
; ShortMonth   -  Returns a three character month name
; LongDay      -  Returns the full name of the day of the week
; ShortDay     -  Returns a thre character day of the week name

UpperCase PROC  FAR USES DS,line:DWORD RETURNS newline:DWORD

        lds     si,line
        les     di,newline
        cld
        lodsb
        stosb
        mov     cl,al
        xor     ch,ch
        jcxz    @@03
@@01:
        lodsb
        cmp     al,'a'
        jb      @@02
        cmp     al,'z'
        ja      @@02
        sub     al,20h
@@02:
        stosb
        loop    @@01
@@03:
        ret

UpperCase ENDP

LowerCase PROC  FAR USES DS,line:DWORD RETURNS newline:DWORD

        lds     si,line
        les     di,newline
        cld
        lodsb
        stosb
        mov     cl,al
        xor     ch,ch
        jcxz    @@13
@@11:
        lodsb
        cmp     al,'A'
        jb      @@12
        cmp     al,'Z'
        ja      @@12
        add     al,20h
@@12:
        stosb
        loop    @@11
@@13:
        ret

LowerCase ENDP

LoCase  PROC    FAR char:BYTE:2

        mov     al,char
        cmp     al,'A'
        jb      @@21
        cmp     al,'Z'
        ja      @@21
        add     al,20h
@@21:
        ret

LoCase  ENDP

Left    PROC    FAR USES DS,line:DWORD,count:BYTE:2 RETURNS newline:DWORD

        lds     si,line         ; get the source line
        les     di,newline      ; get the target line
        xor     ah,ah
        mov     al,count        ; get the size to keep
        cmp     al,[si]         ; is the size greater than the string length?
        jb      @@31            ; no.
        mov     al,[si]         ; yes, use the string length, then
@@31:
        cld
        stosb                   ; save the length
        inc     si              ; adjust the source pointer
        mov     cx,ax           ; we need a counter
        rep     movsb           ; move the string
        ret                     ; all done

Left    ENDP

Right   PROC    FAR USES DS,line:DWORD,count:BYTE:2 RETURNS newline:DWORD

        lds     si,line         ; get the source line
        les     di,newline      ; get the target line
        xor     ah,ah
        lodsb
        sub     al,count
        jnc     @@40
        mov     cl,[si-1]
        jmp     short @@41
@@40:
        add     si,ax
        mov     cl,count
@@41:
        mov     al,cl
        xor     ch,ch
        cld
        stosb                   ; save the length
        rep     movsb           ; move the string
        ret                     ; all done

Right   ENDP

NoLeadSpace PROC FAR  USES DS, line:DWORD RETURNS newline:DWORD

        lds     si,line
        les     di,newline
        cld
        lodsb
        xor     ah,ah
        mov     cx,ax
        mov     al,' '
@@51:
        cmp     al,[si]
        jne     @@52
        inc     si
        loop    @@51
@@52:
        mov     al,cl
        stosb
        rep     movsb
        ret

NoLeadSpace ENDP

NoTrailSpace PROC FAR  USES DS, line:DWORD RETURNS newline:DWORD

        lds     si,line
        les     di,newline
        mov     bl,[si]
        xor     bh,bh
        mov     al,' '
@@61:
        cmp     al,[si+bx]
        jne     @@62
        dec     bx
        jnz     @@61
@@62:
        mov     es:[di],bl
        mov     cx,bx
        inc     di
        inc     si
        cld
        rep     movsb
        ret

NoTrailSpace ENDP

Reverse PROC    FAR USES DS,line:DWORD RETURNS newline:DWORD

        lds     si,line
        les     di,newline
        cld
        lodsb
        stosb
        xor     ah,ah
        add     di,ax
        dec     di
        mov     cx,ax
        jcxz    @@72
@@71:
        lodsb
        mov     es:[di],al
        dec     di
        loop    @@71
@@72:
        ret

Reverse ENDP

LastPos PROC    FAR char:BYTE:2,line:DWORD

        les     di,line
        mov     cl,es:[di]
        xor     ch,ch
        add     di,cx
        mov     al,char
        std
        repne   scasb
        jne     @@01
        inc     cx
@@01:
        mov     ax,cx
        ret

LastPos ENDP

NextPos PROC    FAR char:BYTE:2,line:DWORD,last:BYTE:2

        les     di,line
        cld
        mov     cl,es:[di]
        xor     ch,ch
        mov     bx,cx
        mov     al,last
        xor     ah,ah
        sub     cl,al
        add     di,ax
        inc     di
        mov     al,char
        repne   scasb
        je      @@01
        xor     bx,bx
@@01:
        sub     bx,cx
        inc     bx
        mov     ax,bx
        dec     ax
        ret

NextPos ENDP

PrevPos PROC    FAR char:BYTE:2,line:DWORD,last:BYTE:2

        les     di,line
        mov     cl,last
        xor     ch,ch
        dec     cx
        add     di,cx
        mov     al,char
        std
        repne   scasb
        jne     @@01
        inc     cx
@@01:
        mov     ax,cx
        ret

PrevPos ENDP

PadLeft PROC    FAR USES DS,line:DWORD,char:BYTE:2,amount:BYTE:2 RETURNS newline:DWORD

        lds     si,line         ; get source line pointer
        les     di,newline      ; get target line pointer
        mov     cl,amount       ; get desired string length
        mov     al,[si]         ; get source string length
        sub     cl,al
        jnc     @@91
        xor     cl,cl
        mov     amount,al
@@91:
        mov     al,amount
        cld
        stosb
        mov     al,char
        xor     ch,ch
        rep     stosb
        lodsb
        mov     cl,al
        rep     movsb
        ret

PadLeft ENDP

PadRight PROC   FAR USES DS, line:DWORD,char:BYTE:2,amount:BYTE:2 RETURNS newline:DWORD

        lds     si,line
        les     di,newline
        push    di
        cld
        lodsb
        stosb
        mov     cl,al
        xor     ch,ch
        rep     movsb
        mov     cl,amount
        sub     cl,al
        ja      @@A1
        mov     amount,al
        xor     cl,cl
@@A1:
        mov     al,char
        rep     stosb
        pop     di
        mov     al,amount
        stosb
        ret


PadRight ENDP

DelChar PROC    FAR USES DS,char:BYTE:2,line:DWORD RETURNS newline:DWORD

        lds     si,line
        les     di,newline
        cld
        lodsb
        mov     cl,al
        xor     ch,ch
        xor     bx,bx
        mov     ah,char
@@B1:
        lodsb
        cmp     al,ah
        je      @@B2
        mov     es:[di+bx+1],al
        inc     bx
@@B2:
        loop    @@B1
        mov     es:[di],bl
        ret

DelChar ENDP

ChangeChar PROC FAR USES DS,line:DWORD,tchar:BYTE:2,newchar:BYTE:2 RETURNS newline:DWORD

        lds     si,line
        les     di,newline
        cld
        lodsb
        stosb
        xor     ah,ah
        mov     cx,ax
        mov     ah,tchar
        mov     bl,newchar
@@C1:
        lodsb
        cmp     ah,al
        jne     @@C2
        mov     al,bl
@@C2:
        stosb
        loop    @@C1
        ret

ChangeChar ENDP

Chars   PROC    FAR char:BYTE:2,amount:BYTE RETURNS line:DWORD

        les     di,line
        mov     al,amount
        cld
        stosb
        mov     cl,al
        mov     al,char
        rep     stosb
        ret

Chars   ENDP

GetString PROC  NEAR

        xor     ah,ah
        cld
@@E1:
        lodsb
        dec     cx
        jz      @@E2
        add     si,ax
        or      al,al
        jnz     @@E1
@@E2:
        stosb
        mov     cl,al
        rep     movsb
        ret

GetString ENDP

Lmonthdata DB   7,'January',8,'February',5,'March',5,'April',3,'May'
           DB   4,'June',4,'July',6,'August',9,'September',7,'October'
           DB   8,'November',8,'December',0

SmonthData DB   3,'Jan',3,'Feb',3,'Mar',3,'Apr',3,'May',3,'Jun'
           DB   3,'Jul',3,'Aug',3,'Sep',3,'Oct',3,'Nov',3,'Dec'

LongMonth PROC  FAR USES DS,month:WORD RETURNS line:DWORD

        push    cs
        pop     ds
        lea     si,Lmonthdata
        les     di,line
        mov     cx,month
        call    getstring
        ret

LongMonth ENDP

ShortMonth PROC FAR USES DS,month:WORD RETURNS line:DWORD

        push    cs
        pop     ds
        lea     si,Smonthdata
        les     di,line
        mov     cx,month
        call    getstring
        ret

ShortMonth ENDP

Ldaydata DB     6,'Sunday',6,'Monday',7,'Tuesday',9,'WednesDay'
         DB     8,'Thursday',6,'Friday',8,'Saturday'

Sdaydata DB     3,'Sun',3,'Mon',3,'Tue',3,'Wed',3,'Thu',3,'Fri',3,'Sat'

LongDay PROC  FAR USES DS,day:WORD RETURNS line:DWORD

        push    cs
        pop     ds
        lea     si,ldaydata
        les     di,line
        mov     cx,day
        inc     cx
        call    getstring
        ret

LongDay ENDP

ShortDay PROC   FAR USES DS,day:WORD RETURNS line:DWORD

        push    cs
        pop     ds
        lea     si,sdaydata
        les     di,line
        mov     cx,day
        inc     cx
        call    getstring
        ret

ShortDay ENDP

WordCount PROC  FAR USES DS,Line:DWORD

        lds     si,line
        cld
        lodsb
        xor     ah,ah
        mov     cx,ax           ; number of characters
        xor     bx,bx           ; number of words
@@01:
        jcxz    @@EX
        lodsb
        dec     cx
        cmp     al,' '
        je      @@01
@@02:
        inc     bx
@@03:
        jcxz    @@EX
        lodsb
        dec     cx
        cmp     al,' '
        je      @@01
        jmp     short @@03
@@EX:
        mov     ax,bx
        ret

WordCount ENDP

PosWord PROC    FAR USES DS,line:DWORD,number:WORD

        lds     si,line
        mov     cx,number       ; CX = word counter
        cld
        lodsb
        mov     bl,al           ; BL = number of characters
        xor     bh,bh           ; BH = character position
@@01:
        or      bl,bl
        jz      @@03
        lodsb
        inc     bh
        dec     bl
        cmp     al,' '
        je      @@01
        dec     cx
        jcxz    @@EX
@@02:
        or      bl,bl
        jz      @@03
        lodsb
        inc     bh
        dec     bl
        cmp     al,' '
        jne     @@02
        jmp     short   @@01
@@03:
        xor     bh,bh
@@EX:
        mov     al,bh
        xor     ah,ah
        ret

PosWord ENDP

END
